import logging
from typing import Tuple

import zonefilegen.core
import zonefilegen.parsing

def build_soa_record(soa_dict: dict, serial_number: int) -> zonefilegen.core.ResourceRecord:
    """
    Constructs a special SOA record with a specified serial number
    """
    soa_rec = zonefilegen.core.ResourceRecord()
    soa_rec.name = "@"
    soa_rec.ttl = None
    soa_rec.record_class = 'IN'
    soa_rec.record_type = 'SOA'
    soa_rec.data = f"{soa_dict['mname']} {soa_dict['rname']} ("\
        f"{serial_number} "\
        f"{soa_dict['refresh']} "\
        f"{soa_dict['retry']} " \
        f"{soa_dict['expire']} " \
        f"{soa_dict['negative_caching_ttl']})"
    return soa_rec

def build_reverse_zone(network, ptr_candidates: Tuple, default_ttl: int) -> zonefilegen.core.Zone:
    """
    Checks a set of addresses if they are part of a network and
    include them as PTR records in a reverse zone for that network in such case.
    """
    rev_zone = zonefilegen.core.Zone(zonefilegen.parsing.get_rev_zone_name(network), default_ttl)
    included_ptr_names = set()

    for (name, ttl, addr) in ptr_candidates:
        if addr in network:
            rec = zonefilegen.core.ResourceRecord()
            rec.name = zonefilegen.parsing.get_rev_ptr_name(addr, network.prefixlen)
            rec.record_type = 'PTR'
            rec.record_class = 'IN'
            rec.ttl = ttl
            rec.data = name
            if rec.name not in included_ptr_names:
                included_ptr_names.add(rec.name)
                rev_zone.records.append(rec)
    return rev_zone

def build_fwd_zone(zone_name: str, rrset_dict: dict, default_ttl: int) -> zonefilegen.core.Zone:
    """
    Builds a forward DNS zone from a set of records
    """
    fwd_zone = zonefilegen.core.Zone(zone_name, default_ttl)
    
    for rrset in rrset_dict:
        if rrset['name'].endswith('.'):
            name = rrset['name']
        elif rrset['name'] == '@':
            name = zone_name
        else:
            name = f"{rrset['name']}.{zone_name}"
     
        if type(rrset['data']) is list:
            record_data_arr = rrset['data']
        else:
            record_data_arr = [rrset['data']]
     
        for record_data in record_data_arr:
            record = zonefilegen.core.ResourceRecord()
            record.name = name
            record.record_type = rrset['type']
            if not record.record_type in zonefilegen.core.RECORD_TYPES:
                logging.critical(f"Invalid type in record for {record.name} : {record.record_type}")
                exit(1)
            if 'ttl' in rrset:
                record.ttl = rrset['ttl']
            record.record_class = 'IN'
            record.data = record_data
     
            fwd_zone.records.append(record)
    return fwd_zone

def generate_header(digest: str):
    return f"; Generated by zonefilegen, INPUT_DIGEST: {digest}"


import argparse
import logging
import pathlib
import re

import zonefilegen
import zonefilegen.parsing
import zonefilegen.core

# First line of each generated zone file should be a comment with the 
# SHA-1 hex digest of the input toml file:
# ; Generated by zonefilegen, INPUT_SHA1: ea03443f2d9f8c580e73d2f8cd1016dc7174bddc
FIRST_LINE_PATTERN = re.compile(r'^\s*;.*INPUT_DIGEST:\s+(?P<digest>[0-9a-f]+)')
SOA_PATTERN = re.compile(r'^.+SOA[ \t]+(?P<mname>[\.\w]+)[ \t]+(?P<rname>[\.\w]+)[ \t]+\(\s*(?P<serial>[0-9]+)', re.MULTILINE)

def gen_zone(zone: zonefilegen.core.Zone, output_dir: pathlib.Path, soa_dict: dict, input_digest: str):
    out_filepath: pathlib.Path = output_dir / f"{zone.name}zone"
    logging.info(f"Generating zone file {out_filepath}")
    serial_number = None
    if out_filepath.exists():
        with open(out_filepath, 'r') as f:
            first_line_matches = FIRST_LINE_PATTERN.match(f.readline())
            soa_matches = SOA_PATTERN.search(f.read(), )
            old_digest = None
            old_serial = None
            if first_line_matches:
                old_digest = first_line_matches.group('digest')
            else:
                logging.error(f"Existing zone file {out_filepath} was not generated by this tool. Aborting.") 
                exit(1)

            if soa_matches:
                old_serial = soa_matches.group('serial')
            else:
                logging.warning(f"Didn't find or recognize SOA record in existing zone file {out_filepath}. Serial number will be reset.")
            
            if old_serial and old_digest:
                if old_digest != input_digest:
                    serial_number =  (int(old_serial) + 1) % pow(2, 32)
                    logging.info(f"Changes detected, updating serial to {serial_number}")
                else:
                    serial_number = int(old_serial)
                    logging.info(f"No changes detected, serial remains at {serial_number}")
    
    if serial_number is None:
        serial_number = 1
    soa_rec = zonefilegen.generation.build_soa_record(soa_dict, serial_number)

    with open(out_filepath, 'w') as f:
        f.write(zonefilegen.generation.generate_header(input_digest) + '\n')
        f.write(zone.generate_origin() + '\n')
        f.write(zone.generate_ttl() + '\n')
        f.write(soa_rec.to_line() + '\n')
        for rec in zone.records:
            f.write(rec.to_line() + '\n')

def generate():
    parser = argparse.ArgumentParser(description='Generate DNS zone files.')
    parser.add_argument('input_file', type=pathlib.Path, help='Input file in TOML format')
    parser.add_argument('output_dir', type=pathlib.Path, help='Output directory where generated zone files will be placed')
    parser.add_argument('--verbose', '-v', action='count', default=0)
    parser.add_argument('--version', action='version', version=zonefilegen.__version__)
    args = parser.parse_args()

    log_level = logging.WARNING
    if args.verbose > 1:
        log_level = logging.DEBUG
    elif args.verbose > 0:
        log_level = logging.INFO
    logging.basicConfig(format='%(levelname)s: %(message)s',level=log_level)


    (fwd_zone, reverse_zones, soa_dict, input_digest) = zonefilegen.parsing.parse_toml_file(args.input_file)

    for zone in reverse_zones:
        gen_zone(zone, args.output_dir, soa_dict, input_digest)

    gen_zone(fwd_zone, args.output_dir, soa_dict, input_digest)



from typing import List

RECORD_CLASSES = [
    'IN',
    'CH',
    'HS',
]

RECORD_TYPES =  [
    'A',
    'NS',
    'MD',
    'MF',
    'CNAME',
    'SOA',
    'WKS',
    'PTR',
    'HINFO',
    'MINFO',
    'MX',
    'TXT',
    'RP',
    'AFSDB',
    'X25',
    'ISDN',
    'RT',
    'NSAP',
    'NSAP-PTR',
    'SIG',
    'KEY',
    'PX',
    'GPOS',
    'AAAA',
    'LOC',
    'NXT',
    'EID',
    'NIMLOC',
    'SRV',
    'ATMA',
    'NAPTR',
    'KX',
    'CERT',
    'A6',
    'DNAME',
    'SINK',
    'OPT',
    'APL',
    'DS',
    'SSHFP',
    'IPSECKEY',
    'RRSIG',
    'NSEC',
    'DNSKEY',
    'DHCID',
    'NSEC3',
    'NSEC3PARAM',
    'TLSA',
    'SMIMEA',
    'HIP',
    'NINFO',
    'RKEY',
    'TALINK',
    'CDS',
    'CDNSKEY',
    'OPENPGPKEY',
    'CSYNC',
    'ZONEMD',
    'VCB',
    'SVCB',
    'HTTPS',
    'SPF',
    'UINFO',
    'UID',
    'GID',
    'UNSPEC',
    'NID',
    'L32',
    'L64',
    'LP',
    'EUI48',
    'EUI64',
    'TKEY',
    'TSIG',
    'IXFR',
    'AXFR',
    'MAILB',
    'MAILA',
    '*',
    'URI',
    'CAA',
    'AVC',
    'DOA',
    'AMTRELAY',
    'TA',
    'DLV',
]

class ResourceRecord():
    def __init__(self):
        self.name = None
        self.ttl = None
        self.record_class = None
        self.record_type = None
        self.data = None
    
    def to_line(self):
        ttl_str = str(self.ttl) if self.ttl else ''
        record_class_str = str(self.record_class) if self.record_class else ''
        return f"{self.name} {ttl_str} {record_class_str} {self.record_type} {self.data}"


class Zone():
    def __init__(self, name: str, default_ttl: int):
        self.name = name
        self.default_ttl = default_ttl
        self.records = []
    
    def generate_origin(self):
        return f"$ORIGIN {self.name}"

    def generate_ttl(self):
        return f"$TTL {self.default_ttl}"



import ipaddress
import hashlib
import ipaddress
import logging
import pathlib
from typing import List, Tuple

import toml
import zonefilegen
import zonefilegen.core
import zonefilegen.generation

def get_rev_zone_name(network) -> str:
    """
    Cuts off the first few blocks of a reverse pointer for a network address
    to create a suitable reverse zone name for a certain prefix length.
    """
    if type(network) is ipaddress.IPv4Network:
        divisor = 8
        address_len = 32
    elif type(network) is ipaddress.IPv6Network:
        divisor = 4
        address_len = 128
    else:
        raise Exception(f"Invalid network type: {network}")

    blocks_to_cut = int((address_len - network.prefixlen) / divisor)
    return '.'.join(network.network_address.reverse_pointer.split('.')[blocks_to_cut:None]) + '.'

def get_rev_ptr_name(address, prefix_len) -> str:
    """
    Cuts off the last few blocks of a reverse pointer for an address
    to create a suitable reverse pointer name for a certain prefix length.
    """
    if type(address) is ipaddress.IPv4Address:
        divisor = 8
        address_len = 32
    elif type(address) is ipaddress.IPv6Address:
        divisor = 4
        address_len = 128
    else:
        raise Exception(f"Invalid address type: {address}")

    blocks_to_cut = int(((address_len - prefix_len) / divisor))
    return '.'.join(address.reverse_pointer.split('.')[None:blocks_to_cut])

def parse_toml_file(input_file_path: pathlib.Path) -> Tuple[zonefilegen.core.Zone, List[zonefilegen.core.Zone], dict, str]:
    """
    Parses a toml file with DNS records and generates one forward zone and one or 
    more reverse zones. Additionally, a dict with info about the SOA record and a digest of the source file is returned for embedding 
    in the generated files, to detect when serial number needs to be updated.
    """
    with open(input_file_path, 'r') as f:
        data = toml.load(f)
        f.seek(0)
        file_digest = hashlib.sha1(f.read().encode()).hexdigest()
    
    fwd_zone = zonefilegen.generation.build_fwd_zone(data['origin'], data['rrset'], data['default_ttl'])

    ipv4_ptr_candidates = []
    ipv6_ptr_candidates = []

    for rec in fwd_zone.records:
        if rec.record_type == 'A':
            ipv4_ptr_candidates.append((rec.name, rec.ttl, ipaddress.IPv4Address(rec.data)))
        elif rec.record_type == 'AAAA':
            ipv6_ptr_candidates.append((rec.name, rec.ttl, ipaddress.IPv6Address(rec.data)))

    reverse_zones = []
    for network_str in data['networks']:
        network = ipaddress.ip_network(network_str, strict=True)
        if type(network) is ipaddress.IPv4Network:
            reverse_zones.append(zonefilegen.generation.build_reverse_zone(network, ipv4_ptr_candidates, data['default_ttl']))
            if network.prefixlen % 8 != 0:
                logging.fatal("IPv4 network prefix must be divisible by 8")
                exit(1)

        elif type(network) is ipaddress.IPv6Network:
            reverse_zones.append(zonefilegen.generation.build_reverse_zone(network, ipv6_ptr_candidates, data['default_ttl']))
            if network.prefixlen % 4 != 0:
                logging.fatal("IPv6 network prefix must be divisible by 4")
                exit(1)

    return (fwd_zone, reverse_zones, data['soa'], file_digest)